import getopt
import ConfigParser
import socket
import sys
import time


##
## parse command-line arguments
##

quiet = 0

try:
    (opts, arg) = getopt.getopt(sys.argv[1:], "q")
    if arg:
        raise ValueError
    for opt in opts:
        if opt[0] == "-q":
            quiet = 1
except (getopt.error, ValueError):
    sys.exit("Syntax: %s [-q]" % sys.argv[0])


##
## load configuration
##
config = ConfigParser.ConfigParser({ "connection": "network",
                                     "address":    "192.168.0.1",
                                     "dev":        "ttyS0",
                                     "baud":       "9600", 
                                     "password":   "1234",
                                     "frequency":  "120", })
config.read("rt311up.cfg")


if not ("auth" in config.options("all")):
    sys.exit('The "rt311up.cfg" file must have an auth string set.')
auth = config.get("all", "auth")

password = config.get("all", "password")

frequency = config.getint("all", "frequency")

connection = config.get("all", "connection")
if connection == "network":

    # set parameters based on telnet connection (generally easier)
    import telnetlib
    password_prompt = "Password:"
    Timeout = None

    # for the telnet interface, the logout() need not do anything
    def logout(conn): 
        pass

    # telnet specific parameter - server address (IP or host name)
    server_address = config.get("network", "address")


elif connection == "serial": 
    
    # set parameters based on RS232 connection (occasionally troublesome)
    import seriallib
    password_prompt = "Enter Password : "
    Timeout = seriallib.SerialPortTimeout

    # logout returns to the password prompt from any menu entered by 
    # this program, but it may not if you accidentally leave the 
    # router stuck in some other menu when accessing via a terminal 
    # program!
    def logout(conn):
        conn.write("\033")
        time.sleep(1)
        conn.write("\015exit\015")
        conn.write("\033")
        time.sleep(1)
        conn.write("99\015")

    # serial specific parameters - connection device name and speed
    serial_dev = config.get("serial", "dev")
    serial_baud = config.getint("serial", "baud")

else:
    sys.exit('Invalid connection type "%s" specified in "rt311up.cfg"' % 
             connection)



# handy function to display a timestamped message
def inform(msg):
    now = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time()))
    sys.stderr.write(now + "  " + msg + "\012")



##
## loop forever, sending our uptime periodically to the server
##

rt311 = None
while 1:

    ##
    ## step 0: sleep, and close any existing rt311 connection
    ##
    if rt311: rt311.close()
    time.sleep(frequency)


    ## 
    ## step 1: connect to RT311 command interface
    ##

    if connection == "network":
    
        rt311 = telnetlib.Telnet(server_address)

    elif connection == "serial": 

        try:
            rt311 = seriallib.SerialPort(serial_dev)
        except seriallib.SerialPortLockFileError:
            inform("serial port is locked, aborting attempt")
            continue

        rt311.set_baud(serial_baud)
        rt311.set_character_size(8)
        rt311.set_parity('n')
        rt311.set_stop_bits(1)
        rt311.set_flow_control(0, 0)
        rt311.set_io_timeout(10)
        rt311.write("\015")


    ##
    ## step 2: get pertinent uptime information from RT311
    ##

    try:

        # login
        rt311.read_until(password_prompt)
        rt311.write(password + "\015")

        # select system maintenance
        rt311.read_until("Number:")
        rt311.write("24\015")

        # select command interpreter mode
        rt311.read_until("Number:")
        rt311.write("8\015")
        rt311.read_until("mercury>")

        # send 'sys version' command and read response into variable
        rt311.write("sys version\015")
        info = rt311.read_until("mercury>")

        # send 'sys cpu display' command and read response into variable
        rt311.write("sys cpu display\015")
        cpu_load = rt311.read_until("mercury>")

        # log out
        rt311.write("exit\015")
        rt311.read_until("Number:")
        rt311.write("99\015")

    except Timeout:
        logout(rt311)
        inform("got timeout communicating with RT311, aborting attempt")
        continue



    ##
    ## step 3: parse information
    ##

    import re

    try:
        result = re.search(r" system up time:\s+(\d+):(\d+):\d+", info)
        uptime = (int(result.group(1)) * 60) + int(result.group(2))

        ostype = "Embedded"

        result = re.search(r" RAS version: (\S+) ", info)
        oslevel = "RAS " + result.group(1)

        result = re.search(r"baseline (\d+) ticks", cpu_load)
        hz = int(result.group(1))

        ticks_total = 0
        ticks_used = 0
        ticks_each_second = re.findall(r"\s+\d+\s+(\d+)\s+\d+\.\d+", cpu_load)
        for ticks in ticks_each_second:
            ticks_total = ticks_total + hz
            ticks_used = ticks_used + (hz - int(ticks))
        load = float(ticks_used) / ticks_total

        idle = 0

    except AttributeError:
        inform("problem understanding data returned by RT311, aborting attempt")
        continue


    ##
    ## step 4: build an HTTP message to send to server
    ##

    import urllib
    quote_auth = urllib.quote(auth)
    quote_uptime = urllib.quote("%d" % uptime)
    quote_load = urllib.quote("%.2f" % load)
    quote_idle = urllib.quote("%d" % idle)
    quote_ostype = urllib.quote(ostype)
    quote_oslevel = urllib.quote(oslevel)

    info = "auth=%s&uptime=%s&load=%s&idle=%s&os=%s&oslevel=%s\015\012" % ( 
        quote_auth, quote_uptime, quote_load, 
        quote_idle, quote_ostype, quote_oslevel )

    http_message = ("POST /server.html HTTP/1.1\015\012" +
                    "Host: data.uptimes.net\015\012" +
                    "Connection: Close\015\012" +
                    "User-Agent: upclient/4.2.0/up_rt311-0.01\015\012" +
                    "Content-type: application/x-www-form-urlencoded\015\012" +
                    ("Content-length: %d\015\012" % len(info)) +
                    "\015\012" + 
                    info)


    ##
    ## step 5: send HTTP message to server
    ##

    import select

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(("data.uptimes.net", 80))
    while len(http_message) > 0:
        (read_fds, write_fds, err_fds) = select.select([], 
                                                       [sock.fileno()], 
                                                       [sock.fileno()], 
                                                       60)
        if not (sock.fileno() in write_fds) and not (sock.fileno() in err_fds):
            inform("timeout sending status to data.uptimes.net")
            break
        amt_sent = sock.send(http_message)
        if amt_sent <= 0:
            inform("timeout sending status to data.uptimes.net")
            break
        http_message = http_message[amt_sent:]
    

    ##
    ## step 6: report our progress (if any)
    ##

    if not quiet:
        days = uptime / (24 * 60)
        uptime = uptime % (24 * 60)
        hours = uptime / 60
        minutes = uptime % 60
        inform("sent uptime (%d days, %02d:%02d)" % (days, hours, minutes))

