# 
# library to provide functionality like telnetlib, but for serial ports
#
# note: read_until() timeout slightly different
# note: telnet methods use io_timeout by default - set to None for identical
#       behaviour
# note: could optimize read_until by saving a "checked can't be here" buffer
# note: there is no non-buffering mode for input, which might be nice

import os
import TERMIOS
import termios
import select
import errno

##
## new exceptions that may be raised
##

SerialPortLockFileError = "SerialLockFileError"
SerialPortTimeout       = "SerialPortTimeout"

class SerialPort:

    # set lock directory based on system type (hopefully)
    import stat
    lock_dir = None
    for dirname in ("/var/lock", "/var/spool/lock"):
        try:
            mode = os.stat(dirname)[stat.ST_MODE]
            if stat.S_ISDIR(mode): 
                lock_dir = dirname
                break
        except OSError:
            pass


    def _lock(self):
        import tempfile
        import re

        # write our PID to temporary lock file
        old_tempdir = tempfile.tempdir
        tempfile.tempdir = SerialPort.lock_dir
        tmp_lock_name = tempfile.mktemp()
        tempfile.tempdir = old_tempdir
        tmp_lock = open(tmp_lock_name, "w")
        tmp_lock.write("%d\n" % os.getpid())
        tmp_lock.close()

        try:
            # loop until we successfully lock, or conclusively fail to lock
            while 1:
                # try to create the lock file via link, if it works return
                try:
                    os.link(tmp_lock_name, self.lockfile)
                    return
                except OSError, (no, desc):
                    # any kind of error other than EEXIST is a real error
                    if no != errno.EEXIST: raise

                # at this point, the file exists, so we check for stale locks
                try:
                    chkpid_file = open(self.lockfile, "r")
                except OSError, (no, desc):
                    if no != errno.ENOENT: raise
                    # if file disappeared, try again
                    continue

                # get process id
                pid_str = chkpid_file.readline()
                pid_match = re.search("^\s*(\d+)(?:\s|\Z)", pid_str)
                if pid_match == None: 
                    raise SerialPortLockFileError, 'lockfile missing process id'
                pid = int(pid_match.group(1))

                # check process
                try:
                    os.kill(pid, 0)
                    process_running = 1
                except OSError, (no, desc):
                    if no == errno.ESRCH:
                        process_running = 0
                    else:
                        process_running = 1

                # if process is running, handle it
                if process_running:
                    raise SerialPortLockFileError, 'device locked'

                # otherwise, remove stale lock file and try again
                os.unlink(self.lockfile)

        finally:
            # always delete our temporary file
            try:
                os.unlink(tmp_lock_name)
            except OSError:
                pass


    def __init__(self, port):
        import tty

        # consider us closed until we successfully open and lock the port
        self.closed = 1

        # open TTY (i.e. serial) device
        #  assign to a file object because if the file becomes 
        #  unreferenced, i.e. if there is an error initializing the TTY, 
        #  then the the file descriptor is automatically closed
        self.name = "/dev/%s" % port
        self.fd = os.open(self.name, os.O_RDWR|os.O_NOCTTY|os.O_NDELAY)
        self.file = os.fdopen(self.fd, "r+")

        # get lockfile name and try to lock it
        if SerialPort.lock_dir:
            self.lockfile = "%s/LCK..%s" % (SerialPort.lock_dir, port)
            self._lock()
        else:
            self.lockfile = None
 
        # get TTY attributes
        self.tty_attr = termios.tcgetattr(self.fd)

        # set parameters
        self.set_io_timeout(None)
        self.closed = 0
	self.eof = 0

        # buffer to store input lines
        self.inbuf = ''

	# some stuff to make us look more like a file object
	self.mode = "r+"
	self.softspace = 0

        # set to raw mode
        tty.setraw(self.fd)


    def __del__(self):
        SerialPort.close(self)


    ##
    ## custom methods
    ##
    
    def set_baud(self, baud):
        # convert baud to appropriate value
        tty_baud = getattr(TERMIOS, "B%d" % baud)

        # set the speed
        self.tty_attr[4] = tty_baud   # ispeed
        self.tty_attr[5] = tty_baud   # ospeed
        termios.tcsetattr(self.fd, TERMIOS.TCSADRAIN, self.tty_attr)


    def set_character_size(self, character_size):
        # convert character size to appropriate value
        tty_character_size = getattr(TERMIOS, "CS%d" % character_size)

        # set the character size
        self.tty_attr[2] = self.tty_attr[2] & ~(TERMIOS.CSIZE)       # cflag
        self.tty_attr[2] = self.tty_attr[2] | tty_character_size
        termios.tcsetattr(self.fd, TERMIOS.TCSADRAIN, self.tty_attr)


    def set_parity(self, parity):
        # convert parity to appropriate value
        if (parity == 'n') or (parity == 'N'):
            tty_parity = 0
        elif (parity == 'e') or (parity == 'E'):
            tty_parity = TERMIOS.PARENB
        elif (parity == 'o') or (parity == 'O'):
            tty_parity = TERMIOS.PARENB | TERMIOS.PARODD
        else:
            raise ValueError, 'unknown parity'

        # set the parity
        mask = ~(TERMIOS.PARENB | TERMIOS.PARODD)
        self.tty_attr[2] = (self.tty_attr[2] & mask) | tty_parity   # cflag
        termios.tcsetattr(self.fd, TERMIOS.TCSADRAIN, self.tty_attr)


    def set_stop_bits(self, stop_bits):
        # convert stop bits
        if stop_bits == 1:
            self.tty_attr[2] = self.tty_attr[2] & ~TERMIOS.CSTOPB   # cflag
        elif stop_bits == 2:
            self.tty_attr[2] = self.tty_attr[2] | TERMIOS.CSTOPB    # cflag
        else:
            raise ValueError, 'unknown number of stop bits'

        # set stop bits
        termios.tcsetattr(self.fd, TERMIOS.TCSADRAIN, self.tty_attr)


    def set_flow_control(self, hardware, software):
	if hardware:    # cflag
	    self.tty_attr[2] = self.tty_attr[2] | TERMIOS.CRTSCTS
        else:
	    self.tty_attr[2] = self.tty_attr[2] & ~TERMIOS.CRTSCTS

	if software:    # iflag
	    self.tty_attr[0] = self.tty_attr[0] | (TERMIOS.IXON|TERMIOS.IXOFF)
        else:
	    self.tty_attr[0] = self.tty_attr[0] & ~(TERMIOS.IXON|TERMIOS.IXOFF)

        termios.tcsetattr(self.fd, TERMIOS.TCSANOW, self.tty_attr)


    def set_io_timeout(self, timeout):
        self.io_timeout = timeout


    ##
    ## file methods
    ##

    def close(self):
	# closing twice is a no-op
	if self.closed: return

        # hang up line 
        tty_attr = self.tty_attr
        tty_attr[4] = TERMIOS.B0
        tty_attr[5] = TERMIOS.B0
        termios.tcsetattr(self.fd, TERMIOS.TCSADRAIN, tty_attr)

        # close file info
	# note: closing the file closes the file descriptor - it would
	#       be a possible problem to call os.close() on the file 
	#       descriptor, because in a multi-threaded application this
	#       could cause a reused file desctiptor to be closed
        self.file.close() 
        self.fd = -1

        # remove lockfile (if any)
        if self.lockfile:
            try:
                os.unlink(self.lockfile)
            except OSError:
                pass
            self.lockfile = None

        # and remember that we're closed!
        self.closed = 1

    
    def flush(self):
	# we never buffer output
        pass


    def isatty(self):
        return 0


    def fileno(self):
	# using the fileno() function conveniently raises an error if
	# called on a closed file
        return self.file.fileno()


    # multipurpose read function
    #   expected: string to wait for (may be None)
    #   size: maximum size to read (or -1 for unlimited)
    #   timeout: maximum time between reads (or None for infinity)
    def _read_until(self, expected, size, timeout):
	import string

	# grab any data in our input buffer
        s = self.inbuf
        self.inbuf = ''

        # loop until we reach EOF, timeout, read "size" bytes, or find a match
        while 1:

	    # if we're searching for a string, check for it in our buffer
	    if expected != None:
	        offset = string.find(s, expected)
	        if offset != -1:
                    offset = offset + len(expected)
		    self.inbuf = s[offset:]
		    return s[0:offset]
		
	    # check to see if size was specified
            if size >= 0:
		# if it was, and we've read at least "size" bytes, return
                if len(s) >= size:
                    self.inbuf = s[size:]
                    return s[0:size]

                # otherwise use "size" to set the maximum amount to read
                max_read = size - len(s)

            # if size was not specified, just read a largish chunk
            else:
                max_read = 16384

            # use select to check for input with timeout
            (read_fds, write_fds, err_fds) = select.select([self.fd], 
                                                           [], 
                                                           [self.fd], 
                                                           timeout)

            # if no input, we timed out
            if not (self.fd in read_fds) and not (self.fd in err_fds):
		# save what we've read so far (if anything)
                self.inbuf = s
		# and raise a timeout
                raise SerialPortTimeout, 'timeout on read'

            # at this point, we have data ready to be read
            str_read = os.read(self.fd, max_read)

	    # if we got an empty string, we're at EOF
            if not str_read:
		self.eof = 1
		return s

            # otherwise add to our reply buffer
            s = s + str_read


    def read(self, size=-1):
        return self._read_until(None, size, self.io_timeout)

	
    def readline(self, size=-1):
        return self._read_until("\n", size, self.io_timeout)


    def readlines(self, sizehint=-1):
        if sizehint >= 0:
            s = ''
            while len(s) < sizehint:
                s = s + self.readline()

            
    def write(self, s):
        if self.io_timeout:
            while len(s) > 0:
                (read_fds, write_fds, err_fds) = select.select([], 
                                                               [self.fd], 
                                                               [self.fd], 
                                                               self.io_timeout)
                if (self.fd in write_fds) or (self.fd in err_fds):
                    nwritten = os.write(self.fd, s)
                    s = s[nwritten:]
                else:
                    raise SerialPortTimeout, 'timeout on write'
        else:
            self.file.write(s)
            self.file.flush()

    
    def writelines(self, list):
	for line in list:
	    self.write(line)


    ##
    ## Telnet object methods
    ##

    def read_until(self, expected, timeout=-1):
	if self.closed and self.inbuf == '':
	    raise EOFError
	if timeout < 0:
	    timeout = self.io_timeout
	return self._read_until(expected, -1, timeout)

    def read_all(self):
	return self.read()

    def read_some(self):
	return self.read(1)

    def read_very_eager(self):
	if self.closed and self.inbuf == '':
	    raise EOFError
        s = self.inbuf
        self.inbuf = ''
        (read_fds, write_fds, err_fds) = select.select([self.fd], 
                                         [], 
                                         [self.fd], 
                                         0)
        if (self.fd in read_fds) or (self.fd in err_fds):
            str_read = os.read(self.fd, max_read)
            if not str_read:          # EOF
                pass
            s = s + str_read
        return s

        
    def read_eager(self):
        return self.read_very_eager()
        

    def read_lazy(self): 
	if self.closed and self.inbuf == '':
	    raise EOFError
        s = self.inbuf
	self.inbuf = ""
	return s

    def read_very_lazy(self):
	return self.read_lazy()

